/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DS485_H_
#define DS485_H_

#ifdef __cplusplus
#include <future>
#include <boost/asio/io_service.hpp>
extern "C" {
#endif //__cplusplus

#include "qp_port.h"
#include "dsm.h"
#include <string.h>

#include <digitalSTROM/dsuid.h>
#include <digitalSTROM/ds485-const.h>
#include <digitalSTROM/ds485-socket-server.h>


#define DS485_BROADCAST_ADDRESS DS485_BROADCAST_MASK

/**
 * Get destination address from DS485MessageEvt
 */
#define DS485_GetDestination(evt_)      ((evt_)->message[DS485_DEST_OFFSET] & 0x3F)

/**
 * Get source address from DS485MessageEvt
 */
#define DS485_GetSource(evt_)           ((evt_)->message[DS485_SRC_OFFSET] & 0x3F)

/**
 * Set source address to DS485MessageEvt
 */
#define DS485_SetSource(evt_, src_) do { \
    (evt_)->message[DS485_SRC_OFFSET] &= 0xC0; \
    (evt_)->message[DS485_SRC_OFFSET] |= (src_); \
} while(0);

/**
 * Get payload length from DS485MessageEvt
 */
#define DS485_GetPayloadLength(evt_)    ((evt_)->message[DS485_CL_OFFSET] & (DS485_IsData(evt_) ? 0xFF : 0x0F ) )

/**
 * Get payload length from DS485NetlibPacketMessageEvt
 */
#define DS485_GetNetlibPacketPayloadLength(evt_)    (((evt_)->message[DS485N_PACKET_CL_OFFSET_LOW] << 8) | \
                                                     ((evt_)->message[DS485N_PACKET_CL_OFFSET_HIGH] & 0xFF))

/**
 * Returns true if destination address is broadcast
 */
#define DS485_IsBroadcast(evt_)         (((evt_)->message[DS485_DEST_OFFSET] & DS485_BROADCAST_MASK) != 0)

/**
 * Returns true if message is a data message
 */
#define DS485_IsData(evt_)              (((evt_)->message[DS485_DEST_OFFSET] & DS485_DATACONTROL_MASK) != 0)

/**
 * Returns true if message is a request
 */
#define DS485_IsRequest(evt_)           (((evt_)->message[DS485_SRC_OFFSET] & DS485_SOURCE_MASK) == DS485_REQ_MASK)

/**
 * Returns true if message is a long request
 */
#define DS485_IsNetlibPacketRequest(evt_)   (((evt_)->message[DS485N_PACKET_SRC_OFFSET] & DS485_SOURCE_MASK) == DS485_REQ_MASK)

/**
 * Returns true if message is a response
 */
#define DS485_IsResponse(evt_)          (((evt_)->message[DS485_SRC_OFFSET] & DS485_SOURCE_MASK) == DS485_ACK_RESPONSE_MASK)

/**
 * Returns true if message is an event
 */
#define DS485_IsEvent(evt_)             (((evt_)->message[DS485_SRC_OFFSET] & DS485_SOURCE_MASK) == DS485_EVENT_MASK)

/**
 * Returns DS485CommandType of an event
 */
#define DS485_GetCommandType(evt_)      ((DS485CommandType)((evt_)->message[DS485_SRC_OFFSET] & DS485_SOURCE_MASK))


enum {
    DS485N_PACKET_PAYLOAD_SIZE = DS485N_PACKET_MAX_PAYLOAD_SIZE_RAW,
};

/**
 * DS485 address event structure
 */
typedef struct {
    QEvent super;               /* derives from QEvent */
    uint8_t deviceAddress;      /* dS485 device address */
} DS485AddressEvt;

/**
 * DS485 standard message event structure
 */
typedef struct {
    QEvent super;                               /* derives from QEvent */
    uint16_t size;                              /* message size (including CRC-16) */
    dsuid_t  destinationAddr;
    dsuid_t  sourceAddr;
    uint8_t message[DS485_SHORT_MESSAGE_SIZE];  /* message data (including CRC-16) */
} DS485ShortMessageEvt;

/**
 * DS485 long message event structure
 */
typedef struct {
    QEvent super;                               /* derives from QEvent */
    uint16_t size;                              /* message size (including CRC-16) */
    dsuid_t  destinationAddr;
    dsuid_t  sourceAddr;
    uint8_t message[DS485_LONG_PAYLOAD_SIZE + DS485_HEADER_CRC_SIZE];   /* message data (including CRC-16) */
} DS485MessageEvt;

/**
 * DS485 long netlib packet message event structure
 */
typedef struct {
    QEvent super;                               /* derives from QEvent */
    uint16_t size;                              /* message size (including CRC-16) */
    dsuid_t destinationAddr;
    dsuid_t sourceAddr;
    uint8_t message[DS485N_PACKET_PAYLOAD_SIZE + DS485_HEADER_CRC_SIZE];
} DS485NetlibPacketMessageEvt;

/**
 * DS485 get bus member by index request event.
 * Note: Own bus address is not counted.
 */
typedef struct {
    QEvent super;
    QActive * requester;
    uint8_t index;
} DS485ReqGetBusMemberByIndexEvt;

/**
 * DS485 get bus member by index response event.
 * Returns a valid bus address or DS485_INVALID_ADDRESS.
 */
typedef struct {
    QEvent super;
    uint8_t address;
} DS485ResGetBusMemberByIndexEvt;

/**
 * DS485 emit bus change event.
 */
typedef struct {
    QEvent super;
    dsuid_t dsuid;
    uint8_t flags;
} DS485ReqBusChangeEvt;

/**
 * DS485_EmitBusChange
 * emit signal to ds485 bus to indicate vdc client state changes.
 *
 */
void DS485_EmitBusChange(const char *dsuid, uint8_t flags);


/**
 * Create DS485 control message
 */
DS485MessageEvt * DS485_CreateControlMessage(dsuid_t const *dest, uint8_t src, DS485CommandType type, DS485CommandId cmd, uint8_t payloadLength);

/**
 * Create DS485 data message
 */

DS485MessageEvt * DS485_CreateDataMessageFromId(const dsuid_t *destDsuid, uint16_t sourceId, uint8_t src, DS485CommandType type, uint8_t payloadLength);

/**
 * Create DS485 data message
 */

DS485MessageEvt * DS485_CreateDataMessageFromSrcDestDsuid(const dsuid_t *destDsuid, const dsuid_t *sourceDsuid, uint8_t src, DS485CommandType type, uint8_t payloadLength);

/**
 * Create DS485 long data message
 */
DS485NetlibPacketMessageEvt * DS485_CreateLongDataMessage(dsuid_t const *destDsuid, dsuid_t const *sourceDsuid, uint8_t src, DS485CommandType type, uint16_t payloadLength);

/**
 * Returns next message to transmit or null.
 */
DS485MessageEvt const * DS485_GetNextTxMessage(void);

/**
 * Opaque pointer to DS485 AO
 */
extern QActive * const AO_DS485;

/**
 * Returns margin of the TxMessageQueue
 */
uint8_t DS485_GetTxMessageQueueMargin(void);

/**
 * Returns margin of the deferred TxMessageQueue
 */
uint8_t DS485_GetDeferredTxMessageQueueMargin(void);


#ifdef __cplusplus
}

/**
 * DS485 AO constructor
 */
void DS485_ctor(boost::asio::io_service& ioService, const char* listenPort, bool listenGlobal);

#endif  //__cplusplus

#endif /* DS485_H_ */
