/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>
             Vladimir Burian

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef HEATING_H_
#define HEATING_H_

#include "qp_port.h"
#include "config.h"
#include "dsm.h"
#include "datamodel.h"

//#define HEATING_ENABLED 1
//#if HEATING_ENABLED == 1

typedef enum {
    HEATING_MODE_OFF            = 0,
    HEATING_MODE_PID            = 1,
    HEATING_MODE_FOLLOWER       = 2,
    HEATING_MODE_FIXED          = 3,
    HEATING_MODE_MANUAL         = 4,
    HEATING_LAST_MODE,
} HeatingControllerMode;

typedef enum {
    HEATING_STATE_INTERNAL      = 0,
    HEATING_STATE_EXTERNAL      = 1,
    HEATING_STATE_BACKUP        = 2,
    HEATING_STATE_EMERGENCY     = 3,
    HEATING_LAST_STATE,
    HEATING_INVALID_STATE       = 0xff,
} HeatingControllerState;

typedef enum {
    HEATING_KFW_PID             = 0,    /**< valve set to PID output value */
    HEATING_KFW_FULL_OPEN       = 1,    /**< valve set to full open (maximum) */
} HeatingKeepFloorWarmState;

typedef struct {
    uint16_t t;
    uint16_t tref;
    int16_t e;
    int16_t eprev;
    int32_t i;
    int32_t Yp;
    int32_t Yi;
    int32_t Yd;
    uint8_t Y;
    uint8_t antiWindUp;
    uint8_t padding[2];
} HeatingInternals;


typedef struct {
    QEvent super;
    uint8_t localZoneId;
    uint8_t padding;
} HeatingReqBasicEvt;

typedef struct {
    HeatingReqBasicEvt basic;
    uint8_t heatingState;
    uint8_t padding;
} HeatingReqSetState;

typedef struct {
    HeatingReqBasicEvt basic;
    uint16_t zoneId;
    uint16_t sensorValue;
} HeatingReqFollowSensorPushEvt;

typedef struct {
    HeatingReqBasicEvt basic;
    uint8_t prevControllerMode;
} HeatingReqNotifyConfigUpdateEvt;


typedef struct {
    QStateHandler modeSuperState;       /**< super state for controller modes (off, pid, etc.) */
    uint32_t timeOfStart;               /**< timestamp of controller (pid, external, etc.) start */
    uint32_t kfwTimestamp;              /**< Keep-Floor-Warm: timestamp of last state change */
    uint32_t zfTimestamp;               /**< zone follower: timestamp of the last received control value */
    int32_t i;                          /**< PID: integrator */
    HeatingConfig conf;                 /**< heating configuration */
    uint16_t zfValue;                   /**< zone follower: control value of the zone being followed */
    int16_t e;                          /**< PID: last error */
    int16_t eprev;                      /**< PID: previous error */
    uint8_t antiWindUp;                 /**< PID: anti wind-up state */
    uint8_t heatingState;               /**< state of heating controller (internal, external, etc.) */
    uint8_t kfwState;                   /**< Keep-Floor-Warm: state */
    uint8_t scene;                      /**< Selected scene (operation mode) */
} HeatingController;

typedef struct {
    QHsm super;
    QActive *parent;
    HeatingController *ctrl;
    HeatingConfig *conf;
    uint16_t *operationModes;
    uint16_t zoneId;
    uint8_t localZoneId;
    uint8_t pading[1];
} HeatingZone;

typedef struct {
    QTimeEvt tickEvt;
    QTimeEvtCtr tickInterval;
    uint8_t localId;
    uint8_t countActive;
    uint8_t countActiveDispatched;
    uint8_t padding;
} HeatingTick;

typedef struct {
    QHsm super;
    HeatingZone heatingZone;
    QStateHandler heatingZoneStates[MAX_ZONES];
    HeatingTick heatingTick;
    HeatingController ctrls[MAX_ZONES];
} Heating;

void Heating_ctor(QActive *parent);

void Heating_init();
void Heating_dispatch(QEvent const *e, uint16_t const zoneIds[]);

void Heating_NotifyConfigChange(uint8_t localZoneId);
void Heating_NotifyOperationModesChange(uint8_t localZoneId);
void Heating_UpdateConfig(uint8_t localZoneId, HeatingConfig const *config);
bool Heating_IsConfigurationValid(HeatingConfig const *config, uint8_t zoneId);
bool Heating_IsHeatingStateValid(uint8_t controllerState);
bool Heating_IsControllerRunning(uint8_t localZoneId);
bool Heating_GetInternals(uint8_t localZoneId, HeatingInternals *res);
uint8_t Heating_GetHeatingState(uint8_t localZoneId);
bool Heating_IsControllerModeOff(HeatingConfig const *config);
bool Heating_initializeConfig(HeatingConfig *config);
bool Heating_initializeOperationModes(uint16_t *operationMode, uint16_t length);
//#endif /* HEATING_ENABLED == 1 */
#endif /* HEATING_H_ */
