/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "light.h"
#include "scene.h"
#include "dsm.h"
#include "selectA.h"
#include "datamodel.h"
#include "registers.h"
#include "bsp.h"
#include "bitset8.h"
#include "button.h"
#include "ds485.h"

Q_DEFINE_THIS_MODULE(light)

/* FSM state-handler functions */
static QState Light_initial(Light *me, QEvent const *e);

static QState Light_top(Light *me, QEvent const *e);

static QState Light_special_room_state(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_0(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_1(Light *me, QEvent const *e);

static QState Light_off(Light *me, QEvent const *e);
static QState Light_scene_0(Light *me, QEvent const *e);
static QState Light_scene_0_local(Light *me, QEvent const *e);
static QState Light_local_prio_off_hold(Light *me, QEvent const *e);
static QState Light_standby_hold(Light *me, QEvent const *e);
static QState Light_deep_off_hold(Light *me, QEvent const *e);

static QState Light_on(Light *me, QEvent const *e);
static QState Light_scene_1(Light *me, QEvent const *e);
static QState Light_scene_1_local(Light *me, QEvent const *e);
static QState Light_scene_2_local(Light *me, QEvent const *e);
static QState Light_scene_3_local(Light *me, QEvent const *e);
static QState Light_scene_4_local(Light *me, QEvent const *e);

static QState Light_dimming_down(Light *me, QEvent const *e);
static QState Light_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_0_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_0_dim_up(Light *me, QEvent const *e);
static QState Light_scene_0_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_1_dim_down(Light *me, QEvent const *e);
static QState Light_scene_1_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_1_dim_up(Light *me, QEvent const *e);
static QState Light_scene_1_dimming_up(Light *me, QEvent const *e);
#define GetStateHandlerOffsetFromState(state_)  ((QStateHandler)state_)

static QStateHandler LightOffStates[6];

bool Light_is_off_state(const QStateHandler state) {
    uint8_t i;
    for (i = 0; i < (sizeof(LightOffStates) / sizeof(uint16_t)); ++i) {
        if (state == LightOffStates[i]) {
            return TRUE;
        }
    }
    return FALSE;
}

static void Light_RestartSTOPTimer(Light *me);
static void Light_PostGroupSignal(Light *me, QSignal signal, bool force);
static void Light_StopSTOPTimer(Light *me);
static bool Light_SendSTOP(Light *me);
static bool Light_SendSTOPForGroup(Light *me);
static QTimeEvtCtr Light_HandleDimmingAndGetTimeout(Light *me, Scene scene, OPCOperation operation);

/** Duration of a long downstream message */
#define LONG_DOWNSTREAM_MESSAGE_MS          (480)

/** Duration of a short downstream message */
#define SHORT_DOWNSTREAM_MESSAGE_MS         (240)

/** Duration of device blinking */
#define DEVICE_BLINK_DURATION_MS            (4000)

/* timeout for generating internally a INC/DEC signal */
#define LIGHT_DIM_TIMEOUT_TICKS             (250/BSP_MS_PER_TICKS)

/* timeout for first INC/DEC signal after calling min dim */
#define LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS   (SHORT_DOWNSTREAM_MESSAGE_MS/BSP_MS_PER_TICKS)

/* timeout for generating internally a OPC_INC/OPC_DEC signal */
#define LIGHT_OPC_DIM_TIMEOUT_TICKS         (5000/BSP_MS_PER_TICKS)

/* timeout for ending hold */
#define LIGHT_HOLD_END_TIMEOUT_TICKS        (2100/BSP_MS_PER_TICKS)

/* timeout for setting local priority */
#define LIGHT_SET_LOCAL_PRIO_TIMEOUT_TICKS  (2000/BSP_MS_PER_TICKS)

/* timeout for going from local priority off into standby */
#define LIGHT_STANDBY_TIMEOUT_TICKS         (800/BSP_MS_PER_TICKS)

/* timeout for going from standby into deep off */
#define LIGHT_DEEP_OFF_TIMEOUT_TICKS        (1000/BSP_MS_PER_TICKS)


/*..........................................................................*/
static void Light_RestartSTOPTimer(Light *me)
{
    uint8_t stopTimeout = Datamodel_GetStopTimeoutZoneGroup(me->localZoneId, me->groupId);
    if (stopTimeout != 0) {
        if (me->stopTimer != NULL) {
            /* stop timer */
            QTimeEvt_disarm(&me->stopTimer->super);
        } else {
            me->stopTimer = SelectA_AllocateStopTimer(me->localZoneId, me->groupId, 0);
            if (me->stopTimer == NULL) {
                /* TODO: send error event if no more resources are available */
                return;
            }
        }
        QTimeEvt_postIn(&me->stopTimer->super, me->parent, stopTimeout * 1000 / BSP_MS_PER_TICKS);
    }
}

/*..........................................................................*/
static void Light_StopSTOPTimer(Light *me)
{
    if (me->stopTimer != NULL) {
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
    }
    Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
}

/*..........................................................................*/
static void Light_PostGroupSignal(Light *me, QSignal signal, bool force)
{
    /* no need to post group signals to non standard groups */
    if (me->groupId > NUM_GROUPS_WITH_AREA) {
        return;
    }
    /* Set the localZoneId to 0 to handle the signal in the group state machines.
     * Otherwise signal could be falsely routed in programming mode. */
    SelectA_PostGroupSignal(me->groupId, me->localZoneId, signal, force);
}

/*..........................................................................*/
static bool Light_SendSTOP(Light *me)
{
    bool sendStop = FALSE;

    if (me->stopTimer != NULL){
        /* release timer */
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
        sendStop = TRUE;
    } else if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
        sendStop = TRUE;
    }
    if (sendStop) {
        Scene scene = Scene_GetStopScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
    }
    return sendStop;
}

/*..........................................................................*/
static bool Light_SendSTOPForGroup(Light *me)
{
    if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
        Scene scene = Scene_GetStopScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        return TRUE;
    }
    return FALSE;
}

/*..........................................................................*/
void Light_ctor(Light *me)
{
    QHsm_ctor(&me->super, (QStateHandler)&Light_initial);

    LightOffStates[0] = GetStateHandlerOffsetFromState(&Light_off);
    LightOffStates[1] = GetStateHandlerOffsetFromState(&Light_scene_0);
    LightOffStates[2] = GetStateHandlerOffsetFromState(&Light_scene_0_local);
    LightOffStates[3] = GetStateHandlerOffsetFromState(&Light_local_prio_off_hold);
    LightOffStates[4] = GetStateHandlerOffsetFromState(&Light_standby_hold);
    LightOffStates[5] = GetStateHandlerOffsetFromState(&Light_deep_off_hold);
}

/* HSM definition */

/*..........................................................................*/
static QState Light_initial(Light *me, QEvent const *e)
{
    /* QS dictionary entries */
    QS_FUN_DICTIONARY(&Light_top);

    QS_FUN_DICTIONARY(&Light_off);
    QS_FUN_DICTIONARY(&Light_scene_0);
    QS_FUN_DICTIONARY(&Light_scene_0_local);
    QS_FUN_DICTIONARY(&Light_scene_0_hold);
    QS_FUN_DICTIONARY(&Light_local_prio_off_hold);
    QS_FUN_DICTIONARY(&Light_standby);
    QS_FUN_DICTIONARY(&Light_standby_local);
    QS_FUN_DICTIONARY(&Light_deep_off);
    QS_FUN_DICTIONARY(&Light_deep_off_local);

    QS_FUN_DICTIONARY(&Light_on);
    QS_FUN_DICTIONARY(&Light_scene_1);
    QS_FUN_DICTIONARY(&Light_scene_1_local);
    QS_FUN_DICTIONARY(&Light_scene_2);
    QS_FUN_DICTIONARY(&Light_scene_2_local);
    QS_FUN_DICTIONARY(&Light_scene_3);
    QS_FUN_DICTIONARY(&Light_scene_3_local);
    QS_FUN_DICTIONARY(&Light_scene_4);
    QS_FUN_DICTIONARY(&Light_scene_4_local);

    QS_FUN_DICTIONARY(&Light_dimming_down);
    QS_FUN_DICTIONARY(&Light_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_1_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_1_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_1_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_1_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_2_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_2_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_2_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_2_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_3_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_3_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_3_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_3_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_4_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_4_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_4_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_4_dimming_up);

    return Q_TRAN(&Light_scene_0);
}

/*..........................................................................*/
static QState Light_top(Light *me, QEvent const *e)
{
    /* implements default Light_off behavior */
    switch (e->sig) {
    case Q_INIT_SIG:
        return Q_TRAN(&Light_off);
    case CT_2WAY_UP_TIP_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        return Q_TRAN(&Light_scene_1_local);
    }
    case CT_2WAY_DOWN_TIP_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        return Q_TRAN(&Light_scene_0_local);
    }
    case CT_2WAY_UP_TIP_2X_SIG:
        /* fall through */
    case CT_CLICK_2X_SIG:
        /* fall through */
    case CT_TIP_2X_SIG: {
        return Q_TRAN(&Light_scene_2_local);
    }
    case CT_2WAY_UP_TIP_3X_SIG:
        /* fall through */
    case CT_CLICK_3X_SIG:
        /* fall through */
    case CT_TIP_3X_SIG: {
        return Q_TRAN(&Light_scene_3_local);
    }
    case CT_2WAY_UP_TIP_4X_SIG:
        /* fall through */
    case CT_TIP_4X_SIG: {
        return Q_TRAN(&Light_scene_4_local);
    }
    case BUTTON_GROUP_FORCE_SCENE_0_SIG:
    case CT_2WAY_DOWN_TIP_2X_SIG: {
        /* force local priority off */
        SelectA_ForceCallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, 0, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_TRAN(&Light_scene_0);
    }
    case BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG:
    case BUTTON_GROUP_STANDBY_SIG:
    case CT_2WAY_DOWN_CLICK_3X_SIG:
    case CT_2WAY_DOWN_TIP_3X_SIG: {
        if ((me->stateMachineGroupId != LIGHT_GROUP) || (me->groupId >= USER_GROUP)){
            SelectA_SendGeneralAck();
            return Q_HANDLED();
        }
        if (e->sig == CT_2WAY_DOWN_CLICK_3X_SIG || e->sig == BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG) {
            /* force local priority off */
            SelectA_ForceCallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, 0, me->apartmentWide, me->deviceId);
        }
        /* call STANDBY scene */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, STANDBY, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_TRAN(&Light_scene_0);
    }
    case BUTTON_GROUP_DEEP_OFF_SIG:
    case CT_2WAY_DOWN_TIP_4X_SIG: {
        if ((me->stateMachineGroupId != LIGHT_GROUP) || (me->groupId >= USER_GROUP)) {
            SelectA_SendGeneralAck();
            return Q_HANDLED();
        }
        /* call DEEP OFF scene */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, DEEP_OFF, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_TRAN(&Light_scene_0);
    }
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG: {
        /* send general acknowledge if hold end signal of 2-way button is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_START_SIG: {
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    case DIRECT_SCENE_CALL_INC: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        DsmApi_SendZoneGroupActionCallSceneRequest(sceneZoneGroupEvt->localZoneId,
                                                   sceneZoneGroupEvt->groupId,
                                                   sceneZoneGroupEvt->basic.sceneId,
                                                   sceneZoneGroupEvt->basic.sourceDeviceId,
                                                   sceneZoneGroupEvt->basic.force);
    }
        /* fall through */
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        bool force = sceneZoneGroupEvt->basic.force;
        /* unconditionally call scene locally */
        SelectA_SendSceneZoneGroup(me->localZoneId, me->groupId, scene, force);
        /* fall through */
    }
    case CALL_SCENE_BROADCAST_SIG: {
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if ((sceneBasicEvt->sourceZoneId != me->localZoneId) || (sceneBasicEvt->sourceGroupId != me->groupId)) {
            bool force = sceneBasicEvt->force;
            uint8_t sceneId = sceneBasicEvt->sceneId;
            if (Scene_IsStopScene(sceneId)) {
                Light_StopSTOPTimer(me);
                return Q_HANDLED();
            }
            if (Scene_IsOffScene(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, force);
                return Q_TRAN(&Light_scene_0);
            }
            if (Scene_IsOnScene(sceneId) || Scene_IsScene2(sceneId) || Scene_IsScene3(sceneId) || Scene_IsScene4(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, force);
                return Q_TRAN(&Light_scene_1);
            }
            if (((sceneId == AUTO_STANDBY) || (sceneId == STANDBY) || (sceneId == DEEP_OFF) || (sceneId == SLEEPING) || (sceneId == ABSENT)) &&
                (me->stateMachineGroupId != SHADE_GROUP)) {
                Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
                return Q_TRAN(&Light_scene_0);
            }
            if (sceneId == INC_S) {
                return Q_TRAN(&Light_scene_1_dim_up);
            }
            if (sceneId == DEC_S) {
                return Q_TRAN(&Light_scene_1_dim_down);
            }
        }
        /* don't change state */
        return Q_HANDLED();
    }
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case BUTTON_GROUP_LOCAL_PRIO_OFF_SIG: {
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Light_special_room_state(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_CLICK_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG: {
        return Q_TRAN(&Light_scene_0_local);
    }
    }
    return Q_SUPER(&Light_top);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_0(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_0);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_1(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_1);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_off(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        /* scene 1 */
        return Q_TRAN(&Light_scene_1_local);
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge if short-long, hold start and end signal is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Light_scene_0_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        if (me->stateMachineGroupId == SHADE_GROUP){
            return Q_TRAN(&Light_scene_0_dimming_down);
        } else {
            return Q_TRAN(&Light_local_prio_off_hold);
        }
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_0);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    case DIRECT_SCENE_CALL_INC: {
        return Q_TRAN(&Light_scene_0_dimming_up);
    }
    }
    return Q_SUPER(&Light_top);
}

/*..........................................................................*/
static QState Light_scene_0(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG: {
        /* force local priority off */
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_scene_0_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        Scene scene = Scene_GetOffScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_0);
}

/*..........................................................................*/
static QState Light_local_prio_off_hold(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* force local priority off */
        Scene scene = 0;
        SelectA_ForceCallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        /* arm timers */
        if ((me->stateMachineGroupId == LIGHT_GROUP)  && (me->groupId < USER_GROUP)) {
            /* arm standby timer only for light group */
            Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_STANDBY_TIMEOUT_TICKS, STANDBY_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        }
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_STANDBY_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Light_RestartSTOPTimer(me);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_STANDBY_SIG: {
        return Q_TRAN(&Light_scene_0);
    }
    case STANDBY_SIG: {
        /* call standby scene on all devices in the zone */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, STANDBY, me->apartmentWide, me->deviceId);
        return Q_TRAN(&Light_standby_hold);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_standby_hold(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DEEP_OFF_TIMEOUT_TICKS, DEEP_OFF_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEEP_OFF_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_DEEP_OFF_SIG: {
        return Q_TRAN(&Light_scene_0);
    }
    case DEEP_OFF_SIG: {
        /* call deep off scene on all devices in the zone */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, DEEP_OFF, me->apartmentWide, me->deviceId);
        return Q_TRAN(&Light_deep_off_hold);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_deep_off_hold(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        /* force local priority off */
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_START_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        return Q_TRAN(Light_scene_0);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_on(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        /* off scene */
        return Q_TRAN(&Light_scene_0_local);
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge if hold end signal is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_top);
}

/*..........................................................................*/
static QState Light_scene_1(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_1_dimming_down);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_1_dimming_up);
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_1);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_1_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetOnScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_2_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetScene2(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_3_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetScene3(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_4_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetScene4(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_0_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene DEC or OPC_DEC */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecScene(me->buttonId), OPC_DEC);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        /* call scene DEC (continue) and re-arm timer */
        QTimeEvtCtr dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecContinueScene(me->buttonId), OPC_DEC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_0_dim_up);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_scene_0_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_0_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_0);
}

/*..........................................................................*/
static QState Light_scene_0_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        uint8_t numShortDSMsgs;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call CallSceneMin_A with scene as parameter */
        Scene scene = Scene_GetOnScene(me->buttonId);
        numShortDSMsgs = SelectA_CallSceneMinZoneGroup(me->apartmentWide? 0 : me->localZoneId, me->groupId, scene, me->deviceId);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, numShortDSMsgs * LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        QTimeEvtCtr dimTimeout;
        /* call scene INC (continue) and re-arm timer */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetIncContinueScene(me->buttonId), OPC_INC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_1_dim_down);
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QTimeEvtCtr Light_HandleDimmingAndGetTimeout(Light *me, Scene scene, OPCOperation operation)
{
    QTimeEvtCtr dimTimeout;
    /* call scene or opc operation */
    if (me->outputChannel == 0) {
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_StopSTOPTimer(me);
        dimTimeout = LIGHT_DIM_TIMEOUT_TICKS;
    } else {
        SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, operation, me->outputChannel, me->deviceId);
        dimTimeout = LIGHT_OPC_DIM_TIMEOUT_TICKS;
    }
    return dimTimeout;
}

/*..........................................................................*/
static QState Light_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene DEC or OPC_DEC */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecScene(me->buttonId), OPC_DEC);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        /* call scene DEC (continue) and re-arm timer */
        QTimeEvtCtr dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecContinueScene(me->buttonId), OPC_DEC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene INC or OPC_INC */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetIncScene(me->buttonId), OPC_INC);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        QTimeEvtCtr dimTimeout;
        /* call scene INC (continue) and re-arm timer */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetIncContinueScene(me->buttonId), OPC_INC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_1_dim_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    default:
        break;
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_1_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_1_dim_up);
    }
    }
    return Q_SUPER(&Light_dimming_down);
}

/*..........................................................................*/
static QState Light_scene_1_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_1_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_1_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_1_dim_down);
    }
    }
    return Q_SUPER(&Light_dimming_up);
}
