/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "AsioVdcClient.h"

#include <cassert>

#include "ReconnectingSocket.h"
#include "vdc/VdcFramer.h"
#include "util.h"

#include "ds485.h"
extern "C" {
#include "vdc/vdc-api-proto.h"
}

typedef boost::system::error_code error_code;
typedef boost::asio::ip::tcp::resolver resolver;

using namespace ds::vdsm;

struct AsioVdcClient
{
public:
    AsioVdcClient(boost::asio::io_service& ioService)
        : m_reconnectingSocket(ioService)
        , m_framer(ioService)
        , m_heartbeatTimer(ioService)
    {
        VDSM_ENTER();
        memset(&m_vdcClient, 0, sizeof(m_vdcClient));
        m_vdcClient.asio = this;

        m_reconnectingSocket.connected().connect([this]()
        {
            onConnected();
        });
        m_reconnectingSocket.disconnected().connect([this]()
        {
            onDisconnected();
        });
        m_framer.closed().connect([this]()
        {
            //reconnectingSocket is unable to detect when socket is closed.
            //Trigger explicit reset
            m_reconnectingSocket.reset();
        });
        m_framer.setReadFrameFunction([this](const void* data, size_t size)
        {
            onReadFrame(data, size);
        });
    }

    struct vdc_client* vdcClient() { return &m_vdcClient; }
    ReconnectingSocket& reconnectingSocket() { return m_reconnectingSocket; }
    VdcFramer& framer() { return m_framer; }

private:
    ReconnectingSocket m_reconnectingSocket;
    VdcFramer m_framer;

    boost::asio::basic_waitable_timer<boost::chrono::steady_clock> m_heartbeatTimer;
    struct vdc_client m_vdcClient;

    void onConnected()
    {
        VDSM_ENTER();
        m_framer.start(m_reconnectingSocket.socket());
        asyncHeartbeatTimer();
        vdc_api_send_hello(&m_vdcClient);
    }

    void onDisconnected()
    {
        VDSM_ENTER();
        m_framer.close();
        m_heartbeatTimer.cancel();
        vdc_api_set_vanish_vdc(&m_vdcClient);

        /* autodiscovered devices will be removed by avahi, but manually added
        * devices must not be removed (i.e. we must attempt to reconnect) but
        * we need to reset them completely, because there may be a totally new
        * vDC on the other end later on (i.e. with different dSUID, etc) */
        if (!m_vdcClient.autodiscovered) {
            memset(m_vdcClient.dsuid, 0, DSUID_STR_LEN);
        }
    }

    void onReadFrame(const void* data, size_t size)
    {
        VDSM_ENTER();
        asyncHeartbeatTimer();
        vdc_api_process_message(&m_vdcClient, reinterpret_cast<const unsigned char *>(data), size);
    }

    void asyncHeartbeatTimer()
    {
        m_heartbeatTimer.expires_from_now(m_framer.readTimeout()
            * (200 + randInt() % 100) / 400);
        m_heartbeatTimer.async_wait([=](const error_code& e)
        {
            if (e) {
                return; //timer was aborted
            }
            VDSM_IENTER();
            vdc_api_send_ping(&m_vdcClient);
            //m_heartbeatTimer is reset when we receive any message back
        });
    }
};

namespace {
    //global shared ioService object
    boost::asio::io_service* g_ioService;
}

extern "C" {
vdc_client_t* AsioVdcClient_new()
{
    VDSM_ENTER();
    auto x = new AsioVdcClient(*g_ioService);
    return x->vdcClient();
}

void AsioVdcClient_free(vdc_client_t* vdcClient)
{
    VDSM_ENTER();
    auto x = static_cast<AsioVdcClient *>(vdcClient->asio);
    delete x;
}

void AsioVdcClient_start(vdc_client_t* vdcClient, const char* address, const char* port)
{
    VDSM_ENTER();
    auto x = static_cast<AsioVdcClient *>(vdcClient->asio);
    x->reconnectingSocket().setQuery(boost::asio::ip::tcp::resolver::query(address, port));
    x->reconnectingSocket().start();
}

bool AsioVdcClient_isOpen(vdc_client_t* vdcClient)
{
    VDSM_ENTER();
    auto x = static_cast<AsioVdcClient *>(vdcClient->asio);
    return x->reconnectingSocket().socket().is_open();
}

void AsioVdcClient_reset(vdc_client_t* vdcClient)
{
    VDSM_ENTER();
    auto x = static_cast<AsioVdcClient *>(vdcClient->asio);
    x->reconnectingSocket().reset();
}

void AsioVdcClient_asyncWriteFrame(vdc_client_t* vdcClient, const void* data, size_t size)
{
    VDSM_ENTER();
    auto x = static_cast<AsioVdcClient *>(vdcClient->asio);
    x->framer().asyncWriteFrame(data, size);
}

} //extern "C"

void AsioVdcClient_init(boost::asio::io_service& ioService)
{
    g_ioService = &ioService;
}
