/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Enrique Robledo Arnuncio, for digitalSTROM AG <era@tegulan.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * vDC host whitelist semantics.
 *
 * The whitelist is stored in a sqlite database file.
 *
 * Filename: Specified by the "-w" command line flag.
 *
 * Table name: "autoconnect_whitelist"
 *
 * dsuid column: A string, the dsuid of the vdc host (primary key).
 *
 * noauto column: An integer with boolean semantics. It is just an informative
 * field, reflecting the last observed value of the noauto txt record in the vdc
 * host service description. This will typically be false for ordinary vdc
 * hosts.
 *
 * autoconnect column: An integer with boolean semantics. It specifies whether
 * the vdSM should connect to the vdc host automatically. The value is set by
 * the vdSM the first time a host's announcement is received. This initial value
 * is true if the initial value of noauto is false, and viceversa.
 *
 * "autoconnect" will never be changed by the vdSM after the initialization. It
 * can be modified by the user via the configuration editor in the dss, or by a
 * technician by directly modifying the sqlite database.
 *
 * name column: An string with the Avahi service name label. This is displayed
 * in the whitelist editor GUI for informative purposes. The maximun length
 * of the name is 64 characters (AVAHI_LABEL_MAX).
 */

#ifndef __VDSM_WHITELIST_H__
#define __VDSM_WHITELIST_H__

#include <digitalSTROM/dsuid.h>

#include <uthash.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct vdc_legacy_whitelist_hash {
    char dsuid[DSUID_STR_LEN];
    UT_hash_handle hh;
} vdc_legacy_whitelist_hash_t;

typedef struct vdc_whitelist_fields {
    bool noauto;
    bool autoconnect;
    const char* name;  /* the NULL pointer value corresponds to SQL NULL */
} vdc_whitelist_fields_t;

bool whitelist_init(const char* sqlite_filename);

bool whitelist_exists();

bool whitelist_import_legacy(vdc_legacy_whitelist_hash_t **legacy_whitelist);

/* @brief whitelist_insert_and_get_autoconnect is used to perform an atomic
 * test/set operation on the whitelist when a vdc host announces itself.
 *
 * @dsuid [in] The id of the vdc host.
 *
 * @fields [in] Its three fields are handled as follows:
 *    autoconnect: Is written into the whitelist if the dsuid was not present in
 *        it, otherwise is ignored.
 *    noauto: Is written into the whitelist, replacing the existing value if
 *        present.
 *    name: Is written into the whitelist, replacing the existing value if
 *        present.
 *
 * The caller keeps ownership of the arguments and their members. They can be
 * freed after whitelist_insert_and_get_autoconnect returns.
 *
 * @return the resulting value of autoconnect in the dsuid whitelist entry.
 */
bool whitelist_insert_and_get_autoconnect(const char* dsuid,
                                          const vdc_whitelist_fields_t* fields);

/* @brief Checks the autoconnect fields for a device in the whitelist.
 *
 * @dsuid [in] The id of the vdc host.
 *
 * @return the resulting value of autoconnect in the dsuid whitelist entry,
 *         or false, if no entry exists.
 */
bool whitelist_get_autoconnect(const char* dsuid);


/* @brief Used for testing */
bool whitelist_get(const char* dsuid, vdc_whitelist_fields_t*);
/* @brief Used for testing. Releases the fields set by whitelist_get. */
void whitelist_free_fields(vdc_whitelist_fields_t* fields);

#ifdef __cplusplus
}
#endif


#endif
